DEFAULT_MIN_PASSING_GRADE = 1.0
DEFAULT_MIN_NUM_CREDITS_FULL_TIME = 12.0
DEFAULT_COURSE_CREDIT_CHECK = 12
DEFAULT_FIRST_TERM_OF_YEAR = "FALL"
DEFAULT_CORE_TERMS = {"FALL", "SPRING"}
DEFAULT_COURSE_LEVEL_PATTERN = r"^(?P<course_level>\d)\d{2}(?:[A-Z]{,2})?$"
DEFAULT_PEAK_COVID_TERMS = {
    ("2019-20", "SPRING"),  # Spring 2020
    ("2019-20", "SUMMER"),  # Summer 2020
    ("2020-21", "FALL"),  # Fall 2020
    ("2020-21", "WINTER"),  # Winter 2020/2021
    ("2020-21", "SPRING"),  # Spring 2021
    ("2020-21", "SUMMER"),  # Summer 2021
}

NUM_COURSE_FEATURE_COL_PREFIX = "num_courses"
FRAC_COURSE_FEATURE_COL_PREFIX = "frac_courses"
DUMMY_COURSE_FEATURE_COL_PREFIX = "took"
